/*************************************************************************/
/*                                                                       */
/*                Institut fr maschinelle Sprachverarbeitung            */
/*                     University of Stuttgart, Germany                  */
/*                         Copyright (c) 1998                            */
/*                        All Rights Reserved.                           */
/*                                                                       */
/*  Permission to use, copy, modify, distribute this software and its    */
/*  documentation for research, educational and individual use only, is  */
/*  hereby granted without fee, subject to the following conditions:     */
/*   1. The code must retain the above copyright notice, this list of    */
/*      conditions and the following disclaimer.                         */
/*   2. Any modifications must be clearly marked as such.                */
/*   3. Original authors' names are not deleted.                         */
/*   4. It may not be redistributed in a modified version or within      */
/*      another distribution                                             */
/*  This software may not be used for commercial purposes without        */
/*  specific prior written permission from the authors.                  */
/*                                                                       */
/*  THE UNIVERSITY OF STUTTGART AND THE CONTRIBUTORS TO THIS WORK        */
/*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING      */
/*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT   */
/*  SHALL THE UNIVERSITY OF STUTTGART NOR THE CONTRIBUTORS BE LIABLE     */
/*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES    */
/*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN   */
/*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,          */
/*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF       */
/*  THIS SOFTWARE.                                                       */
/*                                                                       */
/*************************************************************************/
/*             Author :  Mark Breitenbuecher, Gregor Moehler             */
/*                       (moehler@ims.uni-stuttgart.de)                  */
/*             Date   :  June 1998                                       */
/*************************************************************************/
//changes (Mark Mon Jul  6 11:24:38 MDT 1998):
//parse_1_digit returns now LISP-struct rather than char*
//there was a problem with partly invisible nils bei *00 cardinals
//due to this changes  parse_2_digit has to be adapted
//hope there is more consitency mow

#include <festival.h>     //get_c_string u.s.w.
#include "ims_german_text.h"
#include "ims_patternmatcher.h"
#include "ims_name_syls.h"

#define MAXINTLENGTH 24

LISP parse_1_digit(char digit, char endung){
  switch(digit) 
    {
    case '0': 
      return NIL;
      break;
    case '1':
      if (endung==' ') return cons(rintern("ein"),NIL);
      else if (endung=='s') return cons(rintern("eins"),NIL);
      else return cons(rintern("eine"),NIL);
      break;
    case '2': 
      return cons(rintern("zwei"),NIL);
      break;
    case '3': 
      return cons(rintern("drei"),NIL);
      break;
    case '4': 
      return cons(rintern("vier"),NIL);
      break;
    case '5': 
      return cons(rintern("fnf"),NIL);
      break;
    case '6': 
      return cons(rintern("sechs"),NIL);
      break;
    case '7': 
      return cons(rintern("sieben"),NIL);
      break;
    case '8': 
      return cons(rintern("acht"),NIL);
      break;
    case '9': 
      return cons(rintern("neun"),NIL);
    }
  return cons(rintern("unknown"),NIL);
}

LISP  parse_2_digits(char *number, char endung){
  switch (*number)
    {
    case '0': return append(parse_1_digit(*(++number), endung), NIL);
    case '1':switch (*(++number))
      {
      case '0': return cons(rintern("zehn"), NIL);
      case '1': return cons(rintern("elf"), NIL);
      case '2': return cons(rintern("zwlf"), NIL);
      case '3': return cons(rintern("dreizehn"), NIL);
      case '4': return cons(rintern("vierzehn"), NIL);
      case '5': return cons(rintern("fnfzehn"), NIL);
      case '6': return cons(rintern("sechzehn"), NIL);
      case '7': return cons(rintern("siebzehn"), NIL);
      case '8': return cons(rintern("achtzehn"), NIL);
      case '9': return cons(rintern("neunzehn"), NIL);
      }
    case '2': switch (*(++number))
      {
      case '0': return cons(rintern("zwanzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), 
			    cons(rintern("und"), 
				 cons(rintern("zwanzig"), NIL)));
      }
    case '3': switch (*(++number))
      {
      case '0': return cons(rintern("dreiig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("dreiig"), NIL)));
      }
    case '4': switch (*(++number))
      {
      case '0': return cons(rintern("vierzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung),cons( rintern("und"), cons(rintern("vierzig"), NIL)));
      }
    case '5': switch (*(++number))
      {
      case '0': return cons(rintern("fnfzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("fnfzig"), NIL)));
      }
    case '6':switch (*(++number))
      {
      case '0': return cons(rintern("sechzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("sechzig"), NIL)));
      }
    case '7':switch (*(++number))
      {
      case '0': return cons(rintern("siebzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("siebzig"), NIL)));
      }
    case '8':switch (*(++number))
      {
      case '0': return cons(rintern("achtzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("achtzig"), NIL)));
      }
    case '9':switch (*(++number))
      {
      case '0': return cons(rintern("neunzig"), NIL);
      default : return append(parse_1_digit(*(number++), endung), cons(rintern("und"), cons(rintern("neunzig"), NIL)));
      }
    }
  return rintern("unknown");
}

LISP parse_3_digits(char *number, char endung){
  //int anfang;
  switch(*number)
    {
    case '0': return parse_2_digits(++number , endung);
    case '1': return cons(rintern("ein"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '2': return cons(rintern("zwei"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '3': return cons(rintern("drei"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '4': return cons(rintern("vier"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '5': return cons(rintern("fnf"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '6': return cons(rintern("sechs"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '7': return cons(rintern("sieben"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '8': return cons(rintern("acht"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    case '9': return cons(rintern("neun"), 
			  cons(rintern("hundert"), 
			       parse_2_digits(++number, endung)));
    }
  return rintern("unknown");
}//parse_3_digits


LISP german_parse_cardinal(LISP number){
  
  const char *numberstring= get_c_string(number);
  char numberbuffer[MAXINTLENGTH/3][4]; // jeweils dreierblcke aus numberstring 
  char charbuffer[MAXINTLENGTH];  // hier steht der numberstring drin (mit
  //  fhrenden Leerzeichen aufgefllt, damit er ohne
  //  rest in dreierblcke aufgeteilt werden kann)
  //char helpbuffer[4];             // zur Aufbewahrung der Teilstrings
  int length = strlen(numberstring);
  register int i,j;
  int idiv3;
  LISP retval = NIL;       //in dieser Liste steht am Schlu der Zahlenstring          
   
  if (length <= MAXINTLENGTH) {
    if (atoi(numberstring)==0) return cons(rintern("null"),NIL);
    if (atoi(numberstring)==4711) return  cons(rintern("Klnisch"), cons(rintern("Wasser"),NIL));

    switch(length%3){ //numberstring auf durch drei teilbare laenge mit Nullen auffllen
    case 0:
      charbuffer[0]='\0';
      break;
    case 1: 
      charbuffer[0]='0';
      charbuffer[1]='0';
      charbuffer[2]='\0'; 
      length+=2;
      break;
    case 2: 
      charbuffer[0]='0';
      charbuffer[1]='\0';
      length++;
    }
    strcat(charbuffer, numberstring);
    //    printf("\n charbuffer:%s \n length:%i \n", charbuffer, length);
    for(i=0;i<length;i+=3) //charbuffer in dreierbloecke zerteilen,
      {
	for (j=0;j<=2;++j) 
	  {       //und in numberbuffer abspeichern
	    // idiv3=i/3;
	    numberbuffer[idiv3=i/3][j]=charbuffer[((length-3)-i)+j];
	    numberbuffer[idiv3][3]='\0';
	  }

	//	printf("numberbuffer %i :%s \n", i/3, numberbuffer[i/3]); 
      }
	//    printf("qwe\n");
    for (i=((length/3)-1);i>=0;i--){
      if ('0'==numberbuffer[i][0] && 
	  '0'==numberbuffer[i][1] && 
	  '0'==numberbuffer[i][2])
	;
      else
	
	switch(i){
	case 0: if(numberbuffer[i][1]=='0')
	  retval=append(retval, parse_3_digits(numberbuffer[0], 's'));
	else
	  retval=append(retval, parse_3_digits(numberbuffer[0], ' '));
	  break;
	case 1: retval=append(retval, 
			    append(parse_3_digits(numberbuffer[1], ' '), 
				 cons(rintern("tausend"), NIL)));
	  break;
	case 2: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Million"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					   ,cons(rintern("Millionen"), NIL)));
	  break;
	case 3: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') //dreierblock endet auf "01"
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Milliarde"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					   ,cons(rintern("Milliarden"), NIL)));
	  break;
	case 4: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Billion"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					   ,cons(rintern("Billionen"), NIL)));
	  break;
	case 5: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Billiarde"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					   ,cons(rintern("Billiarden"), NIL)));
	  break;
	case 6: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Trillion"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					   ,cons(rintern("Trillionen"), NIL)));
	  break;
	case 7: if(numberbuffer[i][1]=='0' && numberbuffer[i][2]=='1') 
	  retval=append(retval ,append(parse_3_digits(numberbuffer[i], 'e')
					   ,cons(rintern("Trilliarde"), NIL)));
	else 
	  retval=append(retval, append(parse_3_digits(numberbuffer[i], ' ')
					  ,cons(rintern("Trilliarden"), NIL)));
	
	// der, dem weitere Zahlwrter einfallen, kann hier weitermachen...
	// aber nicht vergessen MAXINTLENGTH zu erhoehen!
	}}
    return retval; }
  else return  cons(rintern("unknown"),NIL) ;
} //german_parse_cardinal


//LISP german_parse_digitlist(char *numberstring){
//  if (strlen(numberstring)==0) return nil; 
//  else if (strlen(numberstring)==1) 
//    return cons(rintern(parse_1_digit(numberstring, "")), NIL); 
//  else append(cons(rintern(parse_1_digit(numberstring, "")), NIL),
//		german_parse_digitlist(*numberstring))


	 //  (cond 
	 //   ((eq? (string-length name) 0) (list nil))
	 //   ((eq? (string-length name) 1) (german_parse_1digit(string-car name)))
	 //   (t
	 //    (append (german_parse_1digit(string-car name))
	 //	   (german_parse_digitlist (string-cdr name))))))
/*
(define (german_parse_1char name)
"(german_parse_1char  NAME) Returns a string containing the german word 
for the single char   NAME"
  (cond
   ((string-equal name "0") (list "null"))
   ((string-equal name "1") (list "eins"))
   ((string-equal name "2") (list "zwei"))
   ((string-equal name "3") (list "drei"))
   ((string-equal name "4") (list "vier"))
   ((string-equal name "5") (list "fnf"))
   ((string-equal name "6") (list "sechs"))
   ((string-equal name "7") (list "sieben"))
   ((string-equal name "8") (list "acht"))
   ((string-equal name "9") (list "neun"))
   ((string-equal name ".") (list "Punkt"))
   ((string-equal name ",") (list "Komma"))
   ((string-equal name ":") (list "Doppelpunkt"))
   ((string-equal name ">") (list "grer" "als"))
   ((string-equal name "<") (list "kleiner" "als"))
   ((string-equal name "/") (list "Slsch"))
   ((string-equal name "?") (list "Fragezeichen"))
   ((string-equal name "+") (list "plus"))
   ((string-equal name "-") (list "minus"))
   ((string-equal name "_") (list "Unterstrich"))
   ((string-equal name "=") (list "ist" "gleich"))
   ((string-equal name "$")  (list "Dollar"))
   ((string-equal name "%")  (list "Prozent"))
   ((string-equal name "")  (list "Paragraph"))
   ((string-equal name "#")  (list "Gartenzaun")) 
   ((string-equal name "+")  (list "plus")) 
   ((string-equal name "*")  (list "Stern")) 
   ((string-equal name "-")  (list "minus")) 
   ((string-equal name "~")  (list "Tilde")) 
   ((string-equal name "&")  (list "und")) 
   ((string-equal name "^")  (list "Dach")) 
   ((string-equal name "@")  (list "Klammeraffe")) 
   ((string-equal name "&")  (list "und")) 
   ((string-equal name "|")  (list "senkrechter" "Strich")) 
   ((string-equal name "\\")  (list "bckslsch")) 
   ((string-equal name "{")  (list "geschweifte" "Klammer" "auf"))
   ((string-equal name "}")  (list "geschweifte" "Klammer" "zu"))
   ((string-equal name "]")  (list "eckige" "Klammer" "zu"))
   ((string-equal name "[")  (list "eckige" "Klammer" "auf"))
   ((string-equal name ")")  (list "Klammer" "zu"))
   ((string-equal name "(")  (list "Klammer" "auf"))
   ((string-equal name "^") (list "Dach"))
   ((string-equal name "!") (list "Ausrufezeichen"))
   ((string-equal name "`") (list "Quote"))
   ((string-equal name "'") (list "Backquote"))
   ((string-matches name "[a-zA-Z]") (list name)) 
   (t (list "unknown"))))
  


(define (german_parse_charlist name)
"(german_parse_charlist NAME) return a list of words with the spoken sequence of chars"
  (cond 
   ((eq? (string-length name) 0) (list nil))
   ((eq? (string-length name) 1) (german_parse_1char(string-car name)))
   (t
    (append (german_parse_1char(string-car name))
	   (german_parse_charlist (string-cdr name))))))

*/

void festival_ims_german_text_init(void){
 init_subr_2("mark-string-match",mark_l_matches_match,
 "(mark-string-match ATOM REGEX)\n\
  Returns the substring of ATOM's Printname that matches REGEX\n\
  (longest match from the beginning), nil if it dosen't match.");
 init_subr_2("mark-string-rest",mark_l_matches_rest,
 "(mark-string-rest ATOM REGEX)\n\
  Returns all but the beginning part of ATOM's Printname that matches\n\
  REGEX (The Part that mark-string-match doesn't return");
 init_subr_1("german_parse_cardinal", german_parse_cardinal,
 "(german_parse_cardinal numberstr)\n\
  Returns a list of words, representing the value of the integer stored in \
  numberstr. \"unknown\" if the integer is bigger than \
\"999999999999999999999999\" ");
   festival_def_utt_module ("Name_Syls", FT_Name_Syls, "Names syllables. See example in festival-documentation");
}







